/*
 * sealtutorial.cpp
 *
 *  Created on: 2022-01-17
 *      Author: Ahmet Can Mert
 */

#include "seal/seal.h"

#include <vector>

using namespace std;
using namespace seal;

void sealcomp(int number1, int number2){
    // compute number1 < number2 homomorphically (print 1 if number1 < number2)

    // ---------------------------------------------------------------
    // -------------------------------------------------------- SERVER
    // ---------------------------------------------------------------

    // scheme
    EncryptionParameters parms(scheme_type::bfv);

    // Ring size (n)
    parms.set_poly_modulus_degree(8192);

    // Ciphertext coefficien modulus (q) (NTT-friendly primes)
    parms.set_coeff_modulus(CoeffModulus::BFVDefault(8192));
    
    // Plaintext coefficient modulus (t) (NTT-friendly prime)
    parms.set_plain_modulus(PlainModulus::Batching(8192, 17));

    // Create SEAL context
    SEALContext context(parms);

    // ---------------------------------------------------------------
    // -------------------------------------------------------- CLIENT
    // ---------------------------------------------------------------

    // Key generation
    KeyGenerator keygen(context);

    // Secret key (sk) and public key (pk)
    SecretKey secret_key = keygen.secret_key();
    PublicKey public_key;
    keygen.create_public_key(public_key);

    // Relinearization keys (for relinearization operations)
    RelinKeys relin_keys;
    keygen.create_relin_keys(relin_keys);

    // Galois keys (for rotation)
    GaloisKeys galois_keys;
    keygen.create_galois_keys(galois_keys);

    // Homomorphic Evaluators
    Encryptor encryptor(context, public_key); // encryptor
    Decryptor decryptor(context, secret_key); // decryptor

    // Batch encoder
    BatchEncoder batch_encoder(context);
    size_t slot_count = batch_encoder.slot_count();
    size_t row_size = slot_count / 2;

    // Encode and encrypt inputs
    vector<uint64_t> plain_matrix1(slot_count, 0);
    plain_matrix1[0] = (number1 & 0x1);
    plain_matrix1[1] = ((number1>>1) & 0x1);

    vector<uint64_t> plain_matrix2(slot_count, 0);
    plain_matrix2[0] = (number2 & 0x1);
    plain_matrix2[1] = ((number2>>1) & 0x1);

    Plaintext plain1;
    batch_encoder.encode(plain_matrix1, plain1);
    Plaintext plain2;
    batch_encoder.encode(plain_matrix2, plain2);

    Ciphertext cipher1;
    encryptor.encrypt(plain1, cipher1);
    Ciphertext cipher2;
    encryptor.encrypt(plain2, cipher2);

    // ---------------------------------------------------------------
    // -------------------------------------------------------- SERVER
    // ---------------------------------------------------------------

    // Generate evaluator
    Evaluator evaluator(context); // evaluator (multiplication, addition, rotate ...)
    
    // Perform homomorphic operations
    Ciphertext a_mult_b;
    Ciphertext C_cipher;

    evaluator.multiply(cipher1, cipher2, a_mult_b);      // a*b
    evaluator.relinearize_inplace(a_mult_b, relin_keys);
    evaluator.sub(cipher2, a_mult_b, C_cipher);          // b-a*b --> C_cipher

    Ciphertext a_add_b;
    Ciphertext D_cipher;
    
    evaluator.add(cipher1, cipher2, a_add_b);    // a+b
    evaluator.sub(a_add_b, a_mult_b, D_cipher);  // a+b - a*b
    evaluator.sub(D_cipher, a_mult_b, D_cipher); // a+b - a*b - a*b

    Plaintext plain3;
    vector<uint64_t> plain_matrix3(slot_count, 0);
    plain_matrix3[0] = 1;
    plain_matrix3[1] = 1;
    batch_encoder.encode(plain_matrix3, plain3);

    evaluator.negate_inplace(D_cipher);
    evaluator.add_plain(D_cipher,plain3,D_cipher); // 1 - (a+b - a*b - a*b) --> D_cipher

    // Final step: (C>>1) + C*(D>>1)
    Ciphertext C_cipher_r;
    Ciphertext D_cipher_r;
    Ciphertext result;

    evaluator.rotate_rows(C_cipher,1,galois_keys,C_cipher_r); // C>>1
    evaluator.rotate_rows(D_cipher,1,galois_keys,D_cipher_r); // D>>1

    evaluator.multiply(C_cipher,D_cipher_r,result);   // C*(D>>1)
    evaluator.relinearize_inplace(result,relin_keys);
    evaluator.add(result,C_cipher_r,result);          // (C>>1) + C*(D>>1)

    // Result is at the first slot of decrypted and decoded result
    // [0/1, x, x, x .. ]
    // [  x, x, x, x .. ]

    // ---------------------------------------------------------------
    // -------------------------------------------------------- CLIENT
    // ---------------------------------------------------------------

    // Decrypt and decode the result
    Plaintext plain_result;
    decryptor.decrypt(result,plain_result);

    vector<uint64_t> matrix_result;
    batch_encoder.decode(plain_result, matrix_result);

    cout << number1 << "<" << number2 << " ? : " << matrix_result[0] << endl;

}

int main()
{
    // Try for all input computations and print out results
    int j,k;
    for (j=0; j<4; j++) {
        for(k=0; k<4; k++) {
            sealcomp(j,k);
        }
    }
    
    return 0;
}

