/*
 * sealtutorial.cpp
 *
 *  Created on: 2022-01-16
 *      Author: Ahmet Can Mert
 */

#include "seal/seal.h"

#include <vector>

using namespace std;
using namespace seal;

int main()
{
    /*
    Set scheme and its parameters
    */
    // Scheme (SEAL supports BFV and CKKS schemes)
    EncryptionParameters parms(scheme_type::bfv);

    // Ring size (n)
    size_t poly_modulus_degree = 8192;
    parms.set_poly_modulus_degree(poly_modulus_degree);

    // Ciphertext coefficien modulus (q) (NTT-friendly primes)
    /*
    You can either 1- use default q values provided by SEAL
                   2- or set bit-size of q values yourself 
    */
    parms.set_coeff_modulus(CoeffModulus::BFVDefault(poly_modulus_degree));
    // parms.set_coeff_modulus(CoeffModulus::Create(poly_modulus_degree, { 30,30,30 }));

    // Plaintext coefficient modulus (t) (NTT-friendly prime)
    parms.set_plain_modulus(PlainModulus::Batching(poly_modulus_degree, 20));

    // Create SEAL context
    SEALContext context(parms);

    auto &context_data = *context.key_context_data();

    // Print parameters
    std::cout << "| poly_modulus_degree: " << context_data.parms().poly_modulus_degree() << std::endl;
    std::cout << "| coeff_modulus size: ";
    std::cout << context_data.total_coeff_modulus_bit_count() << " (";
    auto coeff_modulus = context_data.parms().coeff_modulus();
    std::size_t coeff_modulus_size = coeff_modulus.size();
    for (std::size_t i = 0; i < coeff_modulus_size - 1; i++)
    {
        std::cout << coeff_modulus[i].bit_count() << " + ";
    }
    std::cout << coeff_modulus.back().bit_count();
    std::cout << ") bits" << std::endl;
    std::cout << "| plain_modulus size: " << context_data.parms().plain_modulus().bit_count() << std::endl << std::endl;
    
    // Key generation
    KeyGenerator keygen(context);

    // Secret key (sk) and public key (pk)
    SecretKey secret_key = keygen.secret_key();
    PublicKey public_key;
    keygen.create_public_key(public_key);

    // Relinearization keys (for relinearization operations)
    RelinKeys relin_keys;
    keygen.create_relin_keys(relin_keys);

    // Homomorphic Evaluators
    Encryptor encryptor(context, public_key);  // encryptor
    Evaluator evaluator(context);              // evaluator (multiplication, addition, rotate ...)
    Decryptor decryptor(context, secret_key);  // decryptor

    // Batch encoder
    BatchEncoder batch_encoder(context);

    /*
    The total number of batching `slots' equals the poly_modulus_degree, N, and
    these slots are organized into 2-by-(N/2) matrices that can be encrypted and
    computed on. Each slot contains an integer modulo plain_modulus.
    */
    size_t slot_count = batch_encoder.slot_count();
    size_t row_size = slot_count / 2;

    /*
    The matrix plaintext is simply given to BatchEncoder as a flattened vector
    of numbers. The first `row_size' many numbers form the first row, and the
    rest form the second row. Here we create the following matrix:
        [ 0,  1,  2,  3,  0,  0, ...,  0 ]
        [ 4,  5,  6,  7,  0,  0, ...,  0 ]
    */
    
    vector<uint64_t> pod_matrix(slot_count, 0ULL);
    pod_matrix[0] = 0ULL;
    pod_matrix[1] = 1ULL;
    pod_matrix[2] = 2ULL;
    pod_matrix[3] = 3ULL;
    pod_matrix[row_size] = 4ULL;
    pod_matrix[row_size + 1] = 5ULL;
    pod_matrix[row_size + 2] = 6ULL;
    pod_matrix[row_size + 3] = 7ULL;

    // First we use BatchEncoder to encode the matrix into a plaintext polynomial.
    Plaintext plain_matrix;
    cout << "Encoded plain_matrix." << endl;
    batch_encoder.encode(pod_matrix, plain_matrix);

    // Encrypt
    Ciphertext encrypted_matrix;
    cout << "Encrypt plain_matrix to encrypted_matrix." << endl;
    encryptor.encrypt(plain_matrix, encrypted_matrix);
    cout << "    + Noise budget in encrypted_matrix: " << decryptor.invariant_noise_budget(encrypted_matrix) << " bits" << endl;
    
    // Create the second matrix
    vector<uint64_t> pod_matrix2(slot_count, 0ULL);
    pod_matrix2[0] = 8ULL;
    pod_matrix2[1] = 9ULL;
    pod_matrix2[2] = 10ULL;
    pod_matrix2[3] = 11ULL;
    pod_matrix2[row_size] = 12ULL;
    pod_matrix2[row_size + 1] = 13ULL;
    pod_matrix2[row_size + 2] = 14ULL;
    pod_matrix2[row_size + 3] = 15ULL;

    // BatchEncoder
    Plaintext plain_matrix2;
    cout << "Encoded plain_matrix2." << endl;
    batch_encoder.encode(pod_matrix2, plain_matrix2);

    // Encrypt
    Ciphertext encrypted_matrix2;
    cout << "Encrypt plain_matrix2 to encrypted_matrix2." << endl;
    encryptor.encrypt(plain_matrix2, encrypted_matrix2);
    cout << "    + Noise budget in encrypted_matrix2: " << decryptor.invariant_noise_budget(encrypted_matrix2) << " bits" << endl;
    
    // Calculate encrypted_matrix = (encrypted_matrix * encrypted_matrix2) + encrypted_matrix2 and relinearize
    cout << "encrypted_matrix = (encrypted_matrix * encrypted_matrix2) + encrypted_matrix2 and relinearize." << endl;
    evaluator.multiply_inplace(encrypted_matrix, encrypted_matrix2);
    evaluator.add_inplace(encrypted_matrix, encrypted_matrix2);
    evaluator.relinearize_inplace(encrypted_matrix, relin_keys);
    /*
    How much noise budget do we have left?
    */
    cout << "    + Noise budget in result (encrypted_matrix1): " << decryptor.invariant_noise_budget(encrypted_matrix) << " bits" << endl;

    // Decrypt the result
    Plaintext plain_result;
    cout << "Decrypt and decode result." << endl;
    decryptor.decrypt(encrypted_matrix, plain_result);

    // Decode the result
    vector<uint64_t> pod_result;
    batch_encoder.decode(plain_result, pod_result);
    
    // print the calculated result
    cout << endl;
    cout << "pod_result_calculated[0]         : " << pod_result[0] << endl;
    cout << "pod_result_calculated[1]         : " << pod_result[1] << endl;
    cout << "pod_result_calculated[2]         : " << pod_result[2] << endl;
    cout << "pod_result_calculated[3]         : " << pod_result[3] << endl;
    cout << "pod_result_calculated[0+row_size]: " << pod_result[0+row_size] << endl;
    cout << "pod_result_calculated[1+row_size]: " << pod_result[1+row_size] << endl;
    cout << "pod_result_calculated[2+row_size]: " << pod_result[2+row_size] << endl;
    cout << "pod_result_calculated[3+row_size]: " << pod_result[3+row_size] << endl;

    // print the expected result
    cout << endl;
    cout << "pod_result_expected[0]           : " << (pod_matrix[0] * pod_matrix2[0]) + pod_matrix2[0] << endl;
    cout << "pod_result_expected[1]           : " << (pod_matrix[1] * pod_matrix2[1]) + pod_matrix2[1] << endl;
    cout << "pod_result_expected[2]           : " << (pod_matrix[2] * pod_matrix2[2]) + pod_matrix2[2] << endl;
    cout << "pod_result_expected[3]           : " << (pod_matrix[3] * pod_matrix2[3]) + pod_matrix2[3] << endl;
    cout << "pod_result_expected[0+row_size]  : " << (pod_matrix[0+row_size] * pod_matrix2[0+row_size]) + pod_matrix2[0+row_size] << endl;
    cout << "pod_result_expected[1+row_size]  : " << (pod_matrix[1+row_size] * pod_matrix2[1+row_size]) + pod_matrix2[1+row_size] << endl;
    cout << "pod_result_expected[2+row_size]  : " << (pod_matrix[2+row_size] * pod_matrix2[2+row_size]) + pod_matrix2[2+row_size] << endl;
    cout << "pod_result_expected[3+row_size]  : " << (pod_matrix[3+row_size] * pod_matrix2[3+row_size]) + pod_matrix2[3+row_size] << endl;
    
    // Now, we will look at rotate operation
    // which requires Galois keys
    GaloisKeys galois_keys;
    keygen.create_galois_keys(galois_keys);

    // Rotate rows 4 steps right
    cout << endl;
    cout << "Rotate rows 4 steps right." << endl;
    evaluator.rotate_rows_inplace(encrypted_matrix, -4, galois_keys);
    cout << "    + Noise budget after rotation: " << decryptor.invariant_noise_budget(encrypted_matrix) << " bits" << endl;
    decryptor.decrypt(encrypted_matrix, plain_result);
    
    // Rotate columns (swap)
    cout << "Rotate columns." << endl;
    evaluator.rotate_columns_inplace(encrypted_matrix, galois_keys);
    cout << "    + Noise budget after rotation: " << decryptor.invariant_noise_budget(encrypted_matrix) << " bits" << endl;
    
    // Decrypt and decode
    cout << "Decrypt and decode result." << endl;
    decryptor.decrypt(encrypted_matrix, plain_result);
    batch_encoder.decode(plain_result, pod_result);

    // print the result
    cout << endl;
    cout << "pod_result_rotated[0]            : " << pod_result[0] << endl;
    cout << "pod_result_rotated[1]            : " << pod_result[1] << endl;
    cout << "pod_result_rotated[2]            : " << pod_result[2] << endl;
    cout << "pod_result_rotated[3]            : " << pod_result[3] << endl;
    cout << "pod_result_rotated[4]            : " << pod_result[4] << endl;
    cout << "pod_result_rotated[5]            : " << pod_result[5] << endl;
    cout << "pod_result_rotated[6]            : " << pod_result[6] << endl;
    cout << "pod_result_rotated[7]            : " << pod_result[7] << endl;
    cout << "pod_result_rotated[0+row_size]   : " << pod_result[0+row_size] << endl;
    cout << "pod_result_rotated[1+row_size]   : " << pod_result[1+row_size] << endl;
    cout << "pod_result_rotated[2+row_size]   : " << pod_result[2+row_size] << endl;
    cout << "pod_result_rotated[3+row_size]   : " << pod_result[3+row_size] << endl;
    cout << "pod_result_rotated[4+row_size]   : " << pod_result[4+row_size] << endl;
    cout << "pod_result_rotated[5+row_size]   : " << pod_result[5+row_size] << endl;
    cout << "pod_result_rotated[6+row_size]   : " << pod_result[6+row_size] << endl;
    cout << "pod_result_rotated[7+row_size]   : " << pod_result[7+row_size] << endl;

    // * other useful functions
    // evaluator.add_plain_inplace();
    // evaluator.multiply_plain_inplace();
    // ...

    return 0;
}
