/*
 * sealtutorial.cpp
 *
 *  Created on: 2022-01-16
 *      Author: Ahmet Can Mert
 */

#include "seal/seal.h"

#include <vector>

using namespace std;
using namespace seal;

int main()
{
    /*
    Set scheme and its parameters
    */
    // Scheme (SEAL supports BFV and CKKS schemes)
    EncryptionParameters parms(scheme_type::ckks);

    // Ring size (n)
    size_t poly_modulus_degree = 8192;
    parms.set_poly_modulus_degree(poly_modulus_degree);

    // Ciphertext coefficien modulus (q) (NTT-friendly primes)
    /*
    You can either 1- use default q values provided by SEAL
                   2- or set bit-size of q values yourself 
    */
    // parms.set_coeff_modulus(CoeffModulus::BFVDefault(poly_modulus_degree));
    parms.set_coeff_modulus(CoeffModulus::Create(poly_modulus_degree, { 60,40,40,60 }));

    // scale (plaintext data will be multiplied with scale)
    double scale = pow(2.0, 40);

    // Create SEAL context
    SEALContext context(parms);

    auto &context_data = *context.key_context_data();

    // Print parameters
    std::cout << "| poly_modulus_degree: " << context_data.parms().poly_modulus_degree() << std::endl;
    std::cout << "| coeff_modulus size: ";
    std::cout << context_data.total_coeff_modulus_bit_count() << " (";
    auto coeff_modulus = context_data.parms().coeff_modulus();
    std::size_t coeff_modulus_size = coeff_modulus.size();
    for (std::size_t i = 0; i < coeff_modulus_size - 1; i++)
    {
        std::cout << coeff_modulus[i].bit_count() << " + ";
    }
    std::cout << coeff_modulus.back().bit_count();
    std::cout << ") bits" << std::endl << std::endl;

    // Key generation
    KeyGenerator keygen(context);

    // Secret key (sk) and public key (pk)
    SecretKey secret_key = keygen.secret_key();
    PublicKey public_key;
    keygen.create_public_key(public_key);

    // Relinearization keys (for relinearization operations)
    RelinKeys relin_keys;
    keygen.create_relin_keys(relin_keys);

    // Galois Keys
    GaloisKeys galois_keys;
    keygen.create_galois_keys(galois_keys);

    // Homomorphic Evaluators
    Encryptor encryptor(context, public_key);  // encryptor
    Evaluator evaluator(context);              // evaluator (multiplication, addition, rotate ...)
    Decryptor decryptor(context, secret_key);  // decryptor

    // CKKS encoder
    CKKSEncoder encoder(context);
    
    // The total number of batching `slots' equals poly_modulus_degree / 2
    size_t slot_count = encoder.slot_count();
    cout << "Number of slots: " << slot_count << endl;

    // Create a vector and encode it into a plaintext
    cout << "Encoded plain1." << endl;
    vector<double> input{ 0.0, 1.1, 2.2, 3.3 };
    Plaintext plain;
    encoder.encode(input, scale, plain);

    // Encrypt
    cout << "Encrypt plain1 to encrypted1." << endl;
    Ciphertext encrypted;
    encryptor.encrypt(plain, encrypted);
   
    // Create the second vector
    cout << "Encoded plain2." << endl;
    vector<double> input2{ 3.22, 5.4, 1.1, 6.0 };
    Plaintext plain2;
    encoder.encode(input2, scale, plain2);

    // Encrypt
    cout << "Encrypt plain2 to encrypted2." << endl;
    Ciphertext encrypted2;
    encryptor.encrypt(plain2, encrypted2);

    // Multiply and relinearize
    cout << "encrypted3 = encrypted1 * encrypted2" << endl;
    Ciphertext encrypted3;
    evaluator.multiply(encrypted,encrypted2,encrypted3);
    evaluator.relinearize_inplace(encrypted3, relin_keys);
    
    // Rescale 
    cout << "+ Scale of encrypted3 before rescale: " << log2(encrypted3.scale()) << " bits" << endl;
    cout << "Rescale encrypted3." << endl;
    evaluator.rescale_to_next_inplace(encrypted3);
    cout << "+ Scale of encrypted3 after rescale: " << log2(encrypted3.scale()) << " bits" << endl;

    // Decrypt and decode it
    cout << "Decrypt and decode result." << endl;
    Plaintext plain_result;
    decryptor.decrypt(encrypted3, plain_result);
    vector<double> result;
    encoder.decode(plain_result, result);

    // Print calculated result
    cout << endl;
    cout << "result_calculated[0]: " << result[0] << endl;
    cout << "result_calculated[1]: " << result[1] << endl;
    cout << "result_calculated[2]: " << result[2] << endl;
    cout << "result_calculated[3]: " << result[3] << endl;

    // Print expected results
    cout << endl;
    cout << "result_expected[0]  : " << input[0] * input2[0] << endl;
    cout << "result_expected[1]  : " << input[1] * input2[1] << endl;
    cout << "result_expected[2]  : " << input[2] * input2[2] << endl;
    cout << "result_expected[3]  : " << input[3] * input2[3] << endl;

    return 0;
}
